# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import sglang_scheduler_pb2 as sglang__scheduler__pb2

GRPC_GENERATED_VERSION = '1.74.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in sglang_scheduler_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class SglangSchedulerStub(object):
    """Service definition for SGLang scheduler communication
    This protocol bridges the Rust router and Python scheduler
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Generate = channel.unary_stream(
                '/sglang.grpc.scheduler.SglangScheduler/Generate',
                request_serializer=sglang__scheduler__pb2.GenerateRequest.SerializeToString,
                response_deserializer=sglang__scheduler__pb2.GenerateResponse.FromString,
                _registered_method=True)
        self.Embed = channel.unary_unary(
                '/sglang.grpc.scheduler.SglangScheduler/Embed',
                request_serializer=sglang__scheduler__pb2.EmbedRequest.SerializeToString,
                response_deserializer=sglang__scheduler__pb2.EmbedResponse.FromString,
                _registered_method=True)
        self.HealthCheck = channel.unary_unary(
                '/sglang.grpc.scheduler.SglangScheduler/HealthCheck',
                request_serializer=sglang__scheduler__pb2.HealthCheckRequest.SerializeToString,
                response_deserializer=sglang__scheduler__pb2.HealthCheckResponse.FromString,
                _registered_method=True)
        self.Abort = channel.unary_unary(
                '/sglang.grpc.scheduler.SglangScheduler/Abort',
                request_serializer=sglang__scheduler__pb2.AbortRequest.SerializeToString,
                response_deserializer=sglang__scheduler__pb2.AbortResponse.FromString,
                _registered_method=True)


class SglangSchedulerServicer(object):
    """Service definition for SGLang scheduler communication
    This protocol bridges the Rust router and Python scheduler
    """

    def Generate(self, request, context):
        """Submit a generation request (supports streaming)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Embed(self, request, context):
        """Submit an embedding request
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HealthCheck(self, request, context):
        """Health check and metrics
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Abort(self, request, context):
        """Abort a running request
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SglangSchedulerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Generate': grpc.unary_stream_rpc_method_handler(
                    servicer.Generate,
                    request_deserializer=sglang__scheduler__pb2.GenerateRequest.FromString,
                    response_serializer=sglang__scheduler__pb2.GenerateResponse.SerializeToString,
            ),
            'Embed': grpc.unary_unary_rpc_method_handler(
                    servicer.Embed,
                    request_deserializer=sglang__scheduler__pb2.EmbedRequest.FromString,
                    response_serializer=sglang__scheduler__pb2.EmbedResponse.SerializeToString,
            ),
            'HealthCheck': grpc.unary_unary_rpc_method_handler(
                    servicer.HealthCheck,
                    request_deserializer=sglang__scheduler__pb2.HealthCheckRequest.FromString,
                    response_serializer=sglang__scheduler__pb2.HealthCheckResponse.SerializeToString,
            ),
            'Abort': grpc.unary_unary_rpc_method_handler(
                    servicer.Abort,
                    request_deserializer=sglang__scheduler__pb2.AbortRequest.FromString,
                    response_serializer=sglang__scheduler__pb2.AbortResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'sglang.grpc.scheduler.SglangScheduler', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('sglang.grpc.scheduler.SglangScheduler', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class SglangScheduler(object):
    """Service definition for SGLang scheduler communication
    This protocol bridges the Rust router and Python scheduler
    """

    @staticmethod
    def Generate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/sglang.grpc.scheduler.SglangScheduler/Generate',
            sglang__scheduler__pb2.GenerateRequest.SerializeToString,
            sglang__scheduler__pb2.GenerateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Embed(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sglang.grpc.scheduler.SglangScheduler/Embed',
            sglang__scheduler__pb2.EmbedRequest.SerializeToString,
            sglang__scheduler__pb2.EmbedResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def HealthCheck(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sglang.grpc.scheduler.SglangScheduler/HealthCheck',
            sglang__scheduler__pb2.HealthCheckRequest.SerializeToString,
            sglang__scheduler__pb2.HealthCheckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Abort(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sglang.grpc.scheduler.SglangScheduler/Abort',
            sglang__scheduler__pb2.AbortRequest.SerializeToString,
            sglang__scheduler__pb2.AbortResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)
